! $Id: FF08FixedFormForEver.f90 1859 2015-09-11 13:14:02Z ian $
!> @file
!! Defines the FIXEDFORMFOREVER program, and no, I'm not SHOUTING.


!*******************************************************************************
!!
!> A program to fix-form-ify source.
!!
!! Usage:
!!  FF08FixedFormForEver [options] source-file-specs ...
!!
!! Command line options:
!! - -? or --help     Display command line and program usage help, then exit.
!! - --version        Display program version information, then exit.
!! - --list[:] file   Specify a list file - each line from the given file will 
!!                    be added (in addition to any source file specifications 
!!                    provided on the command line) to the list of source file 
!!                    specifications to be processed.  May be specified 
!!                    multiple times.
!!
!! A source-file-spec of '-' directs that source file specifications will also 
!! be read from the console, one file per line.  A line that is zero length 
!! or all blanks terminates the input.  '-' may only be specified once on the 
!! command line.  This feature is useful with operating system shell provided 
!! pipes between programs.
!!
!! For most platforms a source file specification is the same as the name 
!! of a source file.  Some platforms may offer the ability for a source file 
!! specification to contain wildcards that may expand out to zero or more 
!! source file names.  This requires processor capability that is not 
!! described in the F2003 standard and which is not included in default 
!! builds.
!!
!! STOP codes (which may also be the program's exit code):
!! - no code:         Program completed successfully.  This will be reported 
!!                    by most processors back to the operating system as an 
!!                    exit code of zero.  (I did what you asked.)
!! - 2:               FixedFormerizing commenced, but with errors.
!!                    (I know what you want me to do, but I couldn't do it.)
!!                    This stop code is also returned if a particular 
!!                    statement was too long to be represented in fixed 
!!                    form source (in this case output continues but the 
!!                    statement is truncated).
!! - 3:               There was a problem with the command line arguments.  
!!                    (I haven't got a clue what you want to do.)
!!
!! Error messages are written the error unit.  Beautifully formatted fixed 
!! form source is written to the console.

PROGRAM FF08FixedFormForEver
  
  USE CmdLine
  USE Errors
  USE FileUtilities
  USE FixedFormForEverImplementation
  USE Strings
  
  USE, INTRINSIC :: ISO_FORTRAN_ENV, ONLY: OUTPUT_UNIT, ERROR_UNIT
  
  IMPLICIT NONE
  
  !-----------------------------------------------------------------------------
  ! Local constants
  
  ! Version of this program - for the --version option.
  CHARACTER(*), PARAMETER :: program_version  &
      = 'FF08FixedFormForEver version 1.0'
  
  !-----------------------------------------------------------------------------
  ! Local variables
  
  ! Command line option description and parsing results.
  !
  ! In order: help, version, list, free, console-free, console-fixed
  TYPE(CmdLineOption) :: cmd_line_options(6)
  
  ! List of non-option arguments.
  TYPE(String), ALLOCATABLE :: arguments(:)
  
  ! List of file specifications
  TYPE(String), ALLOCATABLE :: file_specs(:)
  
  ! Error message
  CHARACTER(:), ALLOCATABLE :: msg
  
  INTEGER :: stat             ! Error code for ParseCmdLine call.
  INTEGER :: stop_code        ! Stop code from procedure calls
  INTEGER :: i                ! Loop index
  LOGICAL :: output_free      ! Flag for the --free option.
  LOGICAL :: console_fixed    ! Flag for the --console-fixed option.
  LOGICAL :: console_free     ! Flag for the --console-free option.
  
  ! What to read from the console.
  CHARACTER(5) :: console_read
  
  ! Set to .TRUE. if a dash appears as an argument.
  LOGICAL :: got_dash
  
  !*****************************************************************************
  
  !-----------------------------------------------------------------------------
  ! Set up and process command line options.
  
  ! Keep descriptions consistent with the lead-in comments to this program 
  ! unit (hence the indent).  Line arrangements for the DESCRIPTION argument 
  ! are approximately as they'll be seen for 80 column output, by chance.
  
  cmd_line_options(1) = CmdLineOption(  &
      'help',  &
      CHAR_NAME='?',  &
      DESCRIPTION='Display command line and program usage help, then exit.' )
  
  cmd_line_options(2) = CmdLineOption(  &
      'version',  &
      DESCRIPTION='Display program version information, then exit.' )
  
  cmd_line_options(3) = CmdLineOption(  &
      'list',  &
      DESCRIPTION='Specify a list file - each line from the given file will &
                  &be added (in addition to any source file specifications &
                  &provided on the command line) to the list of source file &
                  &specifications to be processed.  May be specified &
                  &multiple times.',  &
      OPTION_TYPE=ioptArg,  &
      CAN_REPEAT=.TRUE. )
  
  cmd_line_options(4) = CmdLineOption(  &
      'free',  &
      DESCRIPTION='Emits a compact free form variant instead.' )
  
  cmd_line_options(5) = CmdLineOption(  &
      'console-free',  &
      DESCRIPTION='Reads free form input source from the console, after &
          &any other files.  May not be used with a file specification of &
          &"-" or the --console-fixed option.' )
  
  cmd_line_options(6) = CmdLineOption(  &
      'console-fixed',  &
      DESCRIPTION='Reads fixed form input source from the console, after &
          &any other files.  May not be used with a file specification of &
          &"-" or the --console-free option.' )
  
  CALL ParseCmdLine(cmd_line_options, arguments, stat, msg)
  
  IF (stat /= 0) THEN
    PRINT "('Error parsing command line:',A)", msg
    CALL usage_message(OUTPUT_UNIT, cmd_line_options)
    STOP 3
  END IF
  
  ! Check for the --help (or -?) option in priority to all others.
  !
  ! If no arguments or options are provided (i.e. no command line 
  ! arguments) then we have nothing to do.  Interpret that as 
  ! a request for help as well.
  IF ( (cmd_line_options(1)%PresentCount > 0)  &
      .OR. (COMMAND_ARGUMENT_COUNT() == 0) ) THEN
    CALL usage_message(OUTPUT_UNIT, cmd_line_options)
    STOP
  END IF
  
  ! Check for the --version option.
  IF (cmd_line_options(2)%PresentCount > 0) THEN
    PRINT "(A)", program_version
    STOP
  END IF
  
  output_free = cmd_line_options(4)%PresentCount /= 0
  console_free = cmd_line_options(5)%PresentCount /= 0
  console_fixed = cmd_line_options(6)%PresentCount /= 0
  
  ! If - is specified on the command line then read file specs from the 
  ! input unit.  Copy other arguments over to the list of file specs.
  got_dash = .FALSE.
  DO i = 1, SIZE(arguments)
    IF (arguments(i)%item == '-') THEN
      IF (got_dash) THEN
        WRITE ( ERROR_UNIT,  &
            "('''-'' appears more than once on the command line.')" )
        STOP 3
      END IF
      got_dash = .TRUE.
      IF (console_free .OR. console_fixed) THEN
        WRITE ( ERROR_UNIT,  &
            "('A file specification of ""-"" cannot be used if the &
              &--console-free or --console-fixed options are present.')" )
        STOP 3
      END IF
      
      CALL AddFromConsole(file_specs, ERROR_UNIT, stop_code)
      IF (stop_code /= 0) CALL do_stop
    ELSE
      CALL Append(file_specs, arguments(i)%item)
    END IF
  END DO
  
  ! If there are any list files specified then expand the list of files 
  ! to accomodate their contents.
  DO i = 1, cmd_line_options(3)%PresentCount
    CALL AddListFile( file_specs, cmd_line_options(3)%Argument(i)%item,  &
      ERROR_UNIT, stop_code )
    IF (stop_code /= 0) CALL do_stop
  END DO
  
  ! Are we reading source from the console?  If so which form.
  IF (console_free .AND. console_fixed) THEN
    WRITE ( ERROR_UNIT, "('The --console-free and --console-fixed options &
        &cannot both be present.')" )
    STOP 3
  ELSE IF (console_free) THEN
    console_read = 'FREE'
  ELSE IF (console_fixed) THEN
    console_read = 'FIXED'
  ELSE
    console_read = ''
  END IF
  
  !-----------------------------------------------------------------------------
  
  CALL Execute( file_specs, console_read, output_free,  &
      OUTPUT_UNIT, ERROR_UNIT, stop_code )
  
  !-----------------------------------------------------------------------------
  ! Cleanup and shut down.
  
  CALL do_stop
  
CONTAINS
  
  ! F2003 doesn't allow an expression as a stop-code (fixed in F2008) so this 
  ! is a work around.
  SUBROUTINE do_stop
    SELECT CASE (stop_code)
    CASE (0)   ! Fall through to end statement.
    CASE (1) ; STOP 1
    CASE (2) ; STOP 2
    CASE DEFAULT;  STOP 3
    END SELECT
  END SUBROUTINE do_stop
  
  !*****************************************************************************
  !!
  !> Prints the command line usage message for the program.  This is 
  !! intended as the handler for '--help' or similar.
  !!
  !! @param[in]     out_unit          Logical unit connected for formatted 
  !! output to receieve the usage message.
  !!
  !! @param[in]     cmd_line_options  Command line options defined for the 
  !! program.
  !!
  !! Try and keep this consistent with the description in the lead-in 
  !! comments for the main program, or users and (more importantly) 
  !! programmers will get confused.
  
  SUBROUTINE usage_message(out_unit, cmd_line_options)
    
    USE CmdLine
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: out_unit
    TYPE(CmdLineOption), INTENT(IN) :: cmd_line_options(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Format specification for a lines worth of message.
    CHARACTER(*), PARAMETER :: fmt_a = "(A)"
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! First argument on command line, which is normally the program name.
    CHARACTER(:), ALLOCATABLE :: program_name
    
    INTEGER :: stat           ! Error code for command line argument query.
    
    !***************************************************************************
    
    WRITE (out_unit, fmt_a) 'Format your fortran source in ...rather compact&
        &... fixed or free form.'
    WRITE (out_unit, "()")
    
    ! Use actual program name in description, in case we've been renamed.
    CALL GetCmdLineArgument(0, program_name, stat)
    IF (stat /= 0) program_name = 'FF08FixedFormForEver'
    
    WRITE (out_unit, fmt_a) 'Usage: ' // program_name  &
        // ' [options] source-file-specs...'
    WRITE (out_unit, "()")
    WRITE (out_unit, fmt_a) 'Valid options are:'
    WRITE (out_unit, "()")
    
    CALL WriteUsageMessage(cmd_line_options, out_unit)
    
    ! Short and simple - we won't elaborate on names versus spec's here.  We 
    ! rely on format reversion to do multiple records for each 80 character 
    ! (approx) line.
    WRITE (out_unit, fmt_a) 'source-file-specs is a list of the names of &
        &the Fortran source files to ', 'parse.  If there is a file named - &
        &then additional file names will be read ', 'from the console (one &
        &per line).'
    WRITE (out_unit, fmt_a) '','The resulting source code is written to the &
        &console.'
    
  END SUBROUTINE usage_message
  
END PROGRAM FF08FixedFormForEver
